#define NOISE 2
#define SPEED_LOW 1
#define SPEED_MED 2
#define SPEED_HIGH 3
#define REPORT_ID_KEYBOARD 1
#define REPORT_ID_MOUSE    2
#define REPORT_ID_GAMEPAD  3

#include "I2Cdev.h"
#include "MPU6050.h"
// Arduino Wire library is required if I2Cdev I2CDEV_ARDUINO_WIRE implementation
// is used in I2Cdev.h
#if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
    #include "Wire.h"
#endif

// class default I2C address is 0x68
// specific I2C addresses may be passed as a parameter here
// AD0 low = 0x68 (default for InvenSense evaluation board)
// AD0 high = 0x69
MPU6050 accelgyro;
//MPU6050 accelgyro(0x69); // <-- use for AD0 high

int16_t ax, ay, az;
int16_t gx, gy, gz;
int16_t dgx, dgy, dgz;
int16_t mouseSpeed = SPEED_MED;
bool setDG=false;
bool isClickedSpeedButton  = false;
bool isFirstClickedLeft = false;
bool isFirstClickedRight = false;


//////////////////////////////////////////////////////////////////////////////
#include <pt.h>
#include <usbdrv.h>
#include "keycodes.h"



// มาโครสำหรับจำลองการหน่วงเวลาใน protothread
#define PT_DELAY(pt,ms,tsVar) \
  tsVar = millis(); \
  PT_WAIT_UNTIL(pt, millis()-tsVar >= (ms));

/////////////////////////////////////////////////////////////////////
// USB report descriptor, สร้างขึ้นจาก HID Descriptor Tool ซึ่ง
// ดาวน์โหลดได้จาก
//    http://www.usb.org/developers/hidpage/dt2_4.zip
//
// หรือใช้ HIDEdit ซึ่งให้บริการฟรีผ่านเว็บที่ http://hidedit.org/
//
// *** ไม่แนะนำให้สร้างเองด้วยมือ ***
/////////////////////////////////////////////////////////////////////
PROGMEM const char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = 
{
  //////////////////////////////////////
  // โครงสร้าง HID report สำหรับเมาส์ 3 ปุ่ม
  //////////////////////////////////////
  0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
  0x0b, 0x02, 0x00, 0x01, 0x00,  // USAGE (Generic Desktop:Mouse)
  0xa1, 0x01,                    // COLLECTION (Application)
  0x0b, 0x01, 0x00, 0x01, 0x00,  //   USAGE (Generic Desktop:Pointer)
  0xa1, 0x00,                    //   COLLECTION (Physical)
  0x85, 0x02,                    //     REPORT_ID (2)
  0x05, 0x09,                    //     USAGE_PAGE (Button)
  0x1b, 0x01, 0x00, 0x09, 0x00,  //     USAGE_MINIMUM (Button:Button 1)
  0x2b, 0x03, 0x00, 0x09, 0x00,  //     USAGE_MAXIMUM (Button:Button 3)
  0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
  0x25, 0x01,                    //     LOGICAL_MAXIMUM (1)
  0x75, 0x01,                    //     REPORT_SIZE (1)
  0x95, 0x03,                    //     REPORT_COUNT (3)
  0x81, 0x02,                    //     INPUT (Data,Var,Abs)
  0x75, 0x05,                    //     REPORT_SIZE (5)
  0x95, 0x01,                    //     REPORT_COUNT (1)
  0x81, 0x03,                    //     INPUT (Cnst,Var,Abs)
  0x05, 0x01,                    //     USAGE_PAGE (Generic Desktop)
  0x0b, 0x30, 0x00, 0x01, 0x00,  //     USAGE (Generic Desktop:X)
  0x0b, 0x31, 0x00, 0x01, 0x00,  //     USAGE (Generic Desktop:Y)
  0x15, 0x81,                    //     LOGICAL_MINIMUM (-127)
  0x25, 0x7f,                    //     LOGICAL_MAXIMUM (127)
  0x75, 0x08,                    //     REPORT_SIZE (8)
  0x95, 0x02,                    //     REPORT_COUNT (2)
  0x81, 0x06,                    //     INPUT (Data,Var,Rel)
  0xc0,                          //     END_COLLECTION
  0xc0,                          // END_COLLECTION

};
struct ReportMouse
{
  /* +----\------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |Byte \ Bit |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
   * +------\----+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  0        |               Report ID = 2                   |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  1        |              Buttons' statuses                |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  2        |                  Delta X                      |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  3        |                  Delta Y                      |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   */
  uint8_t  report_id;
  uint8_t  buttons;
  int8_t   dx;
  int8_t   dy;
};

ReportMouse reportMouse;

// Protothread states
struct pt main_pt;
struct pt moving_pt;

////////////////////////////////////////////////////////////////
// Automatically called by usbpoll() when host makes a request
////////////////////////////////////////////////////////////////
usbMsgLen_t usbFunctionSetup(uchar data[8])
{
  return 0;  /* Return nothing to host for now */
}

void sendMouse(int8_t dx, int8_t dy, uint8_t buttons)
{
  reportMouse.dx = dx;
  reportMouse.dy = dy;
  reportMouse.buttons = buttons;
  usbSetInterrupt((uchar*)&reportMouse, sizeof(reportMouse));
}

PT_THREAD(moving_task(struct pt *pt))
{
  static uint32_t ts = 0;
  static uint8_t i=0;
  static uint8_t j=0;

  PT_BEGIN(pt);

  for (;;)
  {
    if(digitalRead(PIN_PB1)==0)
    {    
      sendMouse(-gz,-gy,1);
      isClickedSpeedButton = false;
      if(isFirstClickedLeft == false)
      {
        isFirstClickedLeft = true;
        PT_DELAY(pt,100,ts);
      }
    }
    else if(digitalRead(PIN_PB2)==0)
    {
      sendMouse(-gz,-gy,2);
      isClickedSpeedButton = false;
      if(isFirstClickedRight == false)
      {
        isFirstClickedRight = true;
        PT_DELAY(pt,100,ts);
      }
    }
    else if(digitalRead(PIN_PB0)==0 && isClickedSpeedButton == false)
    {
      mouseSpeed += 1;
      if(mouseSpeed > SPEED_HIGH)
      {
        mouseSpeed = SPEED_LOW;
      }
      PT_DELAY(pt,200,ts);
      isClickedSpeedButton = true;
      isFirstClickedLeft = false;
      isFirstClickedRight = false;
    }
    else if(digitalRead(PIN_PB3)==0)
    {
      setDG=false;
      isClickedSpeedButton = false;
      isFirstClickedLeft = false;
      isFirstClickedRight = false;
    }
    else
    {
      sendMouse(-gz,-gy,0);
      isClickedSpeedButton = false;
      isFirstClickedLeft = false;
      isFirstClickedRight = false;
    }
    PT_DELAY(pt,20,ts);
  }
  
  PT_END(pt);
}

PT_THREAD(main_task(struct pt *pt))
{
  PT_BEGIN(pt);

  PT_WAIT_UNTIL(pt,usbInterruptIsReady());
  moving_task(&moving_pt);

  PT_END(pt);
}

void setup()
{
  //delay for gyro to be ready
  delay(1000);

  //set pin
  pinMode(PIN_PB0,INPUT_PULLUP);
  pinMode(PIN_PB1,INPUT_PULLUP);
  pinMode(PIN_PB2,INPUT_PULLUP);
  pinMode(PIN_PB3,INPUT_PULLUP);
  pinMode(PIN_PD3,OUTPUT);

  //close led
  digitalWrite(PIN_PD3,HIGH);
  
    // join I2C bus (I2Cdev library doesn't do this automatically)
    #if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
        Wire.begin();
    #elif I2CDEV_IMPLEMENTATION == I2CDEV_BUILTIN_FASTWIRE
        Fastwire::setup(400, true);
    #endif

    // initialize device
    //Serial.println("Initializing I2C devices...");
    accelgyro.initialize();

    // verify connection
    //Serial.println("Testing device connections...");
    //Serial.println(accelgyro.testConnection() ? "MPU6050 connection successful" : "MPU6050 connection failed");
    while(!accelgyro.testConnection()) ;

  // Initialize USB subsystem
  usbInit();
  usbDeviceDisconnect();
  delay(300);
  usbDeviceConnect();
  
  init_peripheral();

  reportMouse.report_id = REPORT_ID_MOUSE;
  reportMouse.dx = 0;
  reportMouse.dy = 0;
  reportMouse.buttons = 0;

  // Initialize tasks
  PT_INIT(&main_pt);
  PT_INIT(&moving_pt);

  //open led when gyroaccel is ready
  digitalWrite(PIN_PD3,LOW);
}

void loop()
{
  accelgyro.getMotion6(&ax, &ay, &az, &gx, &gy, &gz);
  if(setDG==false)
  {
      //set first raw data
      dgx = gx;
      dgy = gy;
      dgz = gz;
      setDG = true;
  }
  else{
      //decrease raw data scale
      gx = ((gx-dgx) / 600) * mouseSpeed;
      gy = ((gy-dgy) / 600) * mouseSpeed;
      gz = ((gz-dgz) / 600) * mouseSpeed;

      //decrease noise
      if(gx<NOISE && gx >-NOISE)gx=0;
      if(gy<NOISE && gy >-NOISE)gy=0;
      if(gz<NOISE && gz >-NOISE)gz=0;
  }
      
  usbPoll();
  main_task(&main_pt);
}
